"""Tests for forward_ref – ref forwarding for function components."""

import importlib
import sys
from types import ModuleType

import wybthon as _wybthon_pkg  # noqa: F401


class _ClassList:
    def __init__(self):
        self._set = set()

    def add(self, name):
        self._set.add(name)

    def remove(self, name):
        self._set.discard(name)

    def contains(self, name):
        return name in self._set


class _Style:
    def __init__(self):
        self._props = {}

    def setProperty(self, name, value):
        self._props[name] = str(value)

    def removeProperty(self, name):
        self._props.pop(name, None)


class _Node:
    def __init__(self, tag=None, text=None):
        self.tag = tag
        self.nodeValue = text
        self._is_text = text is not None
        self.parentNode = None
        self.childNodes = []
        self.attributes = {}
        self.classList = _ClassList()
        self.style = _Style()
        self.value = ""
        self.checked = False

    @property
    def nextSibling(self):
        if self.parentNode is None:
            return None
        try:
            idx = self.parentNode.childNodes.index(self)
        except ValueError:
            return None
        return self.parentNode.childNodes[idx + 1] if idx + 1 < len(self.parentNode.childNodes) else None

    def appendChild(self, node):
        if getattr(node, "parentNode", None) is not None:
            try:
                node.parentNode.childNodes.remove(node)
            except Exception:
                pass
        node.parentNode = self
        self.childNodes.append(node)
        return node

    def insertBefore(self, node, anchor):
        if getattr(node, "parentNode", None) is not None:
            try:
                node.parentNode.childNodes.remove(node)
            except Exception:
                pass
        node.parentNode = self
        if anchor is None:
            self.childNodes.append(node)
            return node
        try:
            idx = self.childNodes.index(anchor)
        except ValueError:
            self.childNodes.append(node)
            return node
        self.childNodes.insert(idx, node)
        return node

    def removeChild(self, node):
        try:
            self.childNodes.remove(node)
            node.parentNode = None
        except ValueError:
            pass
        return node

    def setAttribute(self, name, value):
        self.attributes[name] = str(value)

    def getAttribute(self, name):
        return self.attributes.get(name)

    def removeAttribute(self, name):
        self.attributes.pop(name, None)


class _Document:
    def __init__(self):
        self._listeners = {}

    def createElement(self, tag):
        return _Node(tag=tag)

    def createTextNode(self, text):
        return _Node(text=str(text))

    def addEventListener(self, event_type, proxy):
        self._listeners.setdefault(event_type, set()).add(proxy)

    def removeEventListener(self, event_type, proxy):
        s = self._listeners.get(event_type)
        if s is not None and proxy in s:
            s.remove(proxy)

    def querySelector(self, sel):
        return _Node(tag="div")

    def querySelectorAll(self, sel):
        return []


def _install_stubs():
    saved = {name: sys.modules.get(name) for name in ("js", "pyodide", "pyodide.ffi")}
    js_mod = ModuleType("js")
    js_mod.document = _Document()
    js_mod.fetch = lambda url: None
    sys.modules["js"] = js_mod

    pyodide = ModuleType("pyodide")
    ffi = ModuleType("pyodide.ffi")
    ffi.create_proxy = lambda fn: fn
    sys.modules["pyodide"] = pyodide
    sys.modules["pyodide.ffi"] = ffi
    setattr(pyodide, "ffi", ffi)
    return saved


def _restore(saved):
    for name, mod in saved.items():
        if mod is None:
            sys.modules.pop(name, None)
        else:
            sys.modules[name] = mod


def _reload_modules():
    dom = importlib.import_module("wybthon.dom")
    importlib.reload(dom)
    events = importlib.import_module("wybthon.events")
    importlib.reload(events)
    component = importlib.import_module("wybthon.component")
    importlib.reload(component)
    context = importlib.import_module("wybthon.context")
    importlib.reload(context)
    reactivity = importlib.import_module("wybthon.reactivity")
    importlib.reload(reactivity)
    hooks = importlib.import_module("wybthon.hooks")
    importlib.reload(hooks)
    vdom = importlib.import_module("wybthon.vdom")
    importlib.reload(vdom)
    return vdom, hooks, dom, component


def _collect_texts(node):
    out = []
    if getattr(node, "_is_text", False):
        out.append(node.nodeValue)
    for ch in getattr(node, "childNodes", []):
        out.extend(_collect_texts(ch))
    return out


def test_forward_ref_renders_normally():
    saved = _install_stubs()
    try:
        vdom, hooks, dom, comp_mod = _reload_modules()

        def _render(props, ref):
            return vdom.h("input", {"type": "text", "value": props.get("value", "")})

        FancyInput = comp_mod.forward_ref(_render)
        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(FancyInput, {"value": "hello"}), root)

        inner = root.element.childNodes[0]
        assert inner.tag == "input"
    finally:
        _restore(saved)


def test_forward_ref_passes_ref():
    saved = _install_stubs()
    try:
        vdom, hooks, dom, comp_mod = _reload_modules()

        captured_ref = [None]

        def _render(props, ref):
            captured_ref[0] = ref
            return vdom.h("input", {"type": "text"})

        FancyInput = comp_mod.forward_ref(_render)
        my_ref = hooks.HookRef(None)
        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(FancyInput, {"ref": my_ref}), root)

        assert captured_ref[0] is my_ref
    finally:
        _restore(saved)


def test_forward_ref_none_when_no_ref():
    saved = _install_stubs()
    try:
        vdom, hooks, dom, comp_mod = _reload_modules()

        captured_ref = ["sentinel"]

        def _render(props, ref):
            captured_ref[0] = ref
            return vdom.h("span", {}, "no ref")

        FancySpan = comp_mod.forward_ref(_render)
        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(FancySpan, {"class": "styled"}), root)

        assert captured_ref[0] is None
    finally:
        _restore(saved)


def test_forward_ref_strips_ref_from_props():
    saved = _install_stubs()
    try:
        vdom, hooks, dom, comp_mod = _reload_modules()

        received_props = [None]

        def _render(props, ref):
            received_props[0] = props
            return vdom.h("div", {}, "child")

        Wrapper = comp_mod.forward_ref(_render)
        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(Wrapper, {"ref": "some_ref", "name": "test"}), root)

        assert "ref" not in received_props[0]
        assert received_props[0]["name"] == "test"
    finally:
        _restore(saved)


def test_forward_ref_has_marker():
    from wybthon.component import forward_ref

    Comp = forward_ref(lambda props, ref: None)
    assert getattr(Comp, "_wyb_forward_ref", False) is True



"""Tests for use_reducer hook."""

import importlib
import sys
import time
from types import ModuleType

import wybthon as _wybthon_pkg  # noqa: F401


class _ClassList:
    def __init__(self):
        self._set = set()

    def add(self, name):
        self._set.add(name)

    def remove(self, name):
        self._set.discard(name)

    def contains(self, name):
        return name in self._set


class _Style:
    def __init__(self):
        self._props = {}

    def setProperty(self, name, value):
        self._props[name] = str(value)

    def removeProperty(self, name):
        self._props.pop(name, None)


class _Node:
    def __init__(self, tag=None, text=None):
        self.tag = tag
        self.nodeValue = text
        self._is_text = text is not None
        self.parentNode = None
        self.childNodes = []
        self.attributes = {}
        self.classList = _ClassList()
        self.style = _Style()
        self.value = ""
        self.checked = False

    @property
    def nextSibling(self):
        if self.parentNode is None:
            return None
        try:
            idx = self.parentNode.childNodes.index(self)
        except ValueError:
            return None
        return self.parentNode.childNodes[idx + 1] if idx + 1 < len(self.parentNode.childNodes) else None

    def appendChild(self, node):
        if getattr(node, "parentNode", None) is not None:
            try:
                node.parentNode.childNodes.remove(node)
            except Exception:
                pass
        node.parentNode = self
        self.childNodes.append(node)
        return node

    def insertBefore(self, node, anchor):
        if getattr(node, "parentNode", None) is not None:
            try:
                node.parentNode.childNodes.remove(node)
            except Exception:
                pass
        node.parentNode = self
        if anchor is None:
            self.childNodes.append(node)
            return node
        try:
            idx = self.childNodes.index(anchor)
        except ValueError:
            self.childNodes.append(node)
            return node
        self.childNodes.insert(idx, node)
        return node

    def removeChild(self, node):
        try:
            self.childNodes.remove(node)
            node.parentNode = None
        except ValueError:
            pass
        return node

    def setAttribute(self, name, value):
        self.attributes[name] = str(value)

    def getAttribute(self, name):
        return self.attributes.get(name)

    def removeAttribute(self, name):
        self.attributes.pop(name, None)


class _Document:
    def __init__(self):
        self._listeners = {}

    def createElement(self, tag):
        return _Node(tag=tag)

    def createTextNode(self, text):
        return _Node(text=str(text))

    def addEventListener(self, event_type, proxy):
        self._listeners.setdefault(event_type, set()).add(proxy)

    def removeEventListener(self, event_type, proxy):
        s = self._listeners.get(event_type)
        if s is not None and proxy in s:
            s.remove(proxy)

    def querySelector(self, sel):
        return _Node(tag="div")

    def querySelectorAll(self, sel):
        return []


def _install_stubs():
    saved = {name: sys.modules.get(name) for name in ("js", "pyodide", "pyodide.ffi")}
    js_mod = ModuleType("js")
    js_mod.document = _Document()
    js_mod.fetch = lambda url: None
    sys.modules["js"] = js_mod

    pyodide = ModuleType("pyodide")
    ffi = ModuleType("pyodide.ffi")
    ffi.create_proxy = lambda fn: fn
    sys.modules["pyodide"] = pyodide
    sys.modules["pyodide.ffi"] = ffi
    setattr(pyodide, "ffi", ffi)
    return saved


def _restore(saved):
    for name, mod in saved.items():
        if mod is None:
            sys.modules.pop(name, None)
        else:
            sys.modules[name] = mod


def _reload_modules():
    dom = importlib.import_module("wybthon.dom")
    importlib.reload(dom)
    events = importlib.import_module("wybthon.events")
    importlib.reload(events)
    component = importlib.import_module("wybthon.component")
    importlib.reload(component)
    context = importlib.import_module("wybthon.context")
    importlib.reload(context)
    reactivity = importlib.import_module("wybthon.reactivity")
    importlib.reload(reactivity)
    hooks = importlib.import_module("wybthon.hooks")
    importlib.reload(hooks)
    vdom = importlib.import_module("wybthon.vdom")
    importlib.reload(vdom)
    return vdom, hooks, dom


def _collect_texts(node):
    out = []
    if getattr(node, "_is_text", False):
        out.append(node.nodeValue)
    for ch in getattr(node, "childNodes", []):
        out.extend(_collect_texts(ch))
    return out


def _counter_reducer(state, action):
    if action == "increment":
        return state + 1
    if action == "decrement":
        return state - 1
    if action == "reset":
        return 0
    return state


def test_use_reducer_initial_render():
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        render_log = []

        def Counter(props):
            count, dispatch = hooks.use_reducer(_counter_reducer, 0)
            render_log.append(count)
            return vdom.h("p", {}, f"Count: {count}")

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(Counter, {}), root)

        assert render_log == [0]
        assert "Count: 0" in _collect_texts(root.element)
    finally:
        _restore(saved)


def test_use_reducer_dispatch():
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        dispatch_ref = [None]
        render_log = []

        def Counter(props):
            count, dispatch = hooks.use_reducer(_counter_reducer, 0)
            dispatch_ref[0] = dispatch
            render_log.append(count)
            return vdom.h("p", {}, f"Count: {count}")

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(Counter, {}), root)
        assert render_log == [0]

        dispatch_ref[0]("increment")
        time.sleep(0.05)
        assert render_log[-1] == 1

        dispatch_ref[0]("increment")
        time.sleep(0.05)
        assert render_log[-1] == 2

        dispatch_ref[0]("decrement")
        time.sleep(0.05)
        assert render_log[-1] == 1

        dispatch_ref[0]("reset")
        time.sleep(0.05)
        assert render_log[-1] == 0
    finally:
        _restore(saved)


def test_use_reducer_with_init():
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        init_calls = [0]

        def init_fn(arg):
            init_calls[0] += 1
            return {"count": arg, "step": 1}

        def reducer(state, action):
            if action == "step":
                return {**state, "count": state["count"] + state["step"]}
            return state

        dispatch_ref = [None]
        state_log = []

        def StepCounter(props):
            state, dispatch = hooks.use_reducer(reducer, 10, init=init_fn)
            dispatch_ref[0] = dispatch
            state_log.append(state)
            return vdom.h("p", {}, f"count={state['count']}")

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(StepCounter, {}), root)

        assert init_calls[0] == 1
        assert state_log[0] == {"count": 10, "step": 1}
        assert "count=10" in _collect_texts(root.element)

        dispatch_ref[0]("step")
        time.sleep(0.05)
        assert state_log[-1] == {"count": 11, "step": 1}
    finally:
        _restore(saved)


def test_use_reducer_dispatch_stable_identity():
    """The dispatch function should be the same reference across renders."""
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        dispatch_ids = []
        setter_ref = [None]

        def MyComp(props):
            _count, dispatch = hooks.use_reducer(_counter_reducer, 0)
            # Also have a useState to trigger re-renders independently
            val, setter = hooks.use_state(0)
            setter_ref[0] = setter
            dispatch_ids.append(id(dispatch))
            return vdom.h("p", {}, "ok")

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(MyComp, {}), root)

        setter_ref[0](1)
        time.sleep(0.05)

        assert len(dispatch_ids) == 2
        assert dispatch_ids[0] == dispatch_ids[1]
    finally:
        _restore(saved)



"""Tests for memo() – memoized function components."""

import importlib
import sys
import time
from types import ModuleType

import wybthon as _wybthon_pkg  # noqa: F401


class _ClassList:
    def __init__(self):
        self._set = set()

    def add(self, name):
        self._set.add(name)

    def remove(self, name):
        self._set.discard(name)

    def contains(self, name):
        return name in self._set


class _Style:
    def __init__(self):
        self._props = {}

    def setProperty(self, name, value):
        self._props[name] = str(value)

    def removeProperty(self, name):
        self._props.pop(name, None)


class _Node:
    def __init__(self, tag=None, text=None):
        self.tag = tag
        self.nodeValue = text
        self._is_text = text is not None
        self.parentNode = None
        self.childNodes = []
        self.attributes = {}
        self.classList = _ClassList()
        self.style = _Style()
        self.value = ""
        self.checked = False

    @property
    def nextSibling(self):
        if self.parentNode is None:
            return None
        try:
            idx = self.parentNode.childNodes.index(self)
        except ValueError:
            return None
        return self.parentNode.childNodes[idx + 1] if idx + 1 < len(self.parentNode.childNodes) else None

    def appendChild(self, node):
        if getattr(node, "parentNode", None) is not None:
            try:
                node.parentNode.childNodes.remove(node)
            except Exception:
                pass
        node.parentNode = self
        self.childNodes.append(node)
        return node

    def insertBefore(self, node, anchor):
        if getattr(node, "parentNode", None) is not None:
            try:
                node.parentNode.childNodes.remove(node)
            except Exception:
                pass
        node.parentNode = self
        if anchor is None:
            self.childNodes.append(node)
            return node
        try:
            idx = self.childNodes.index(anchor)
        except ValueError:
            self.childNodes.append(node)
            return node
        self.childNodes.insert(idx, node)
        return node

    def removeChild(self, node):
        try:
            self.childNodes.remove(node)
            node.parentNode = None
        except ValueError:
            pass
        return node

    def setAttribute(self, name, value):
        self.attributes[name] = str(value)

    def getAttribute(self, name):
        return self.attributes.get(name)

    def removeAttribute(self, name):
        self.attributes.pop(name, None)


class _Document:
    def __init__(self):
        self._listeners = {}

    def createElement(self, tag):
        return _Node(tag=tag)

    def createTextNode(self, text):
        return _Node(text=str(text))

    def addEventListener(self, event_type, proxy):
        self._listeners.setdefault(event_type, set()).add(proxy)

    def removeEventListener(self, event_type, proxy):
        s = self._listeners.get(event_type)
        if s is not None and proxy in s:
            s.remove(proxy)

    def querySelector(self, sel):
        return _Node(tag="div")

    def querySelectorAll(self, sel):
        return []


def _install_stubs():
    saved = {name: sys.modules.get(name) for name in ("js", "pyodide", "pyodide.ffi")}
    js_mod = ModuleType("js")
    js_mod.document = _Document()
    js_mod.fetch = lambda url: None
    sys.modules["js"] = js_mod

    pyodide = ModuleType("pyodide")
    ffi = ModuleType("pyodide.ffi")
    ffi.create_proxy = lambda fn: fn
    sys.modules["pyodide"] = pyodide
    sys.modules["pyodide.ffi"] = ffi
    setattr(pyodide, "ffi", ffi)
    return saved


def _restore(saved):
    for name, mod in saved.items():
        if mod is None:
            sys.modules.pop(name, None)
        else:
            sys.modules[name] = mod


def _reload_modules():
    dom = importlib.import_module("wybthon.dom")
    importlib.reload(dom)
    events = importlib.import_module("wybthon.events")
    importlib.reload(events)
    component = importlib.import_module("wybthon.component")
    importlib.reload(component)
    context = importlib.import_module("wybthon.context")
    importlib.reload(context)
    reactivity = importlib.import_module("wybthon.reactivity")
    importlib.reload(reactivity)
    hooks = importlib.import_module("wybthon.hooks")
    importlib.reload(hooks)
    vdom = importlib.import_module("wybthon.vdom")
    importlib.reload(vdom)
    return vdom, hooks, dom


def _collect_texts(node):
    out = []
    if getattr(node, "_is_text", False):
        out.append(node.nodeValue)
    for ch in getattr(node, "childNodes", []):
        out.extend(_collect_texts(ch))
    return out


def test_memo_renders_initially():
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        render_count = [0]

        def Display(props):
            render_count[0] += 1
            return vdom.h("p", {}, f"value={props.get('value')}")

        MemoDisplay = vdom.memo(Display)

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(MemoDisplay, {"value": "A"}), root)

        assert render_count[0] == 1
        assert "value=A" in _collect_texts(root.element)
    finally:
        _restore(saved)


def test_memo_skips_rerender_same_props():
    """When parent re-renders but passes the same prop references, memo skips."""
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        child_renders = [0]
        parent_setter = [None]
        stable_value = "stable"

        def Child(props):
            child_renders[0] += 1
            return vdom.h("span", {}, f"child:{props.get('label')}")

        MemoChild = vdom.memo(Child)

        def Parent(props):
            count, set_count = hooks.use_state(0)
            parent_setter[0] = set_count
            return vdom.h("div", {}, vdom.h("p", {}, str(count)), vdom.h(MemoChild, {"label": stable_value}))

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(Parent, {}), root)
        assert child_renders[0] == 1

        parent_setter[0](1)
        time.sleep(0.05)

        # Parent re-rendered but memo child should NOT have re-rendered
        assert child_renders[0] == 1
    finally:
        _restore(saved)


def test_memo_rerenders_on_changed_props():
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        child_renders = [0]
        parent_setter = [None]

        def Child(props):
            child_renders[0] += 1
            return vdom.h("span", {}, f"count={props.get('count')}")

        MemoChild = vdom.memo(Child)

        def Parent(props):
            count, set_count = hooks.use_state(0)
            parent_setter[0] = set_count
            return vdom.h("div", {}, vdom.h(MemoChild, {"count": count}))

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(Parent, {}), root)
        assert child_renders[0] == 1
        assert "count=0" in _collect_texts(root.element)

        parent_setter[0](1)
        time.sleep(0.05)

        # Props changed, memo child SHOULD re-render
        assert child_renders[0] == 2
        assert "count=1" in _collect_texts(root.element)
    finally:
        _restore(saved)


def test_memo_custom_comparison():
    """Custom are_props_equal can control when re-renders happen."""
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        child_renders = [0]
        parent_setter = [None]

        def Child(props):
            child_renders[0] += 1
            return vdom.h("span", {}, f"v={props.get('value')}")

        # Custom comparator ignores all differences
        MemoChild = vdom.memo(Child, are_props_equal=lambda old, new: True)

        def Parent(props):
            count, set_count = hooks.use_state(0)
            parent_setter[0] = set_count
            return vdom.h("div", {}, vdom.h(MemoChild, {"value": count}))

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(Parent, {}), root)
        assert child_renders[0] == 1

        parent_setter[0](1)
        time.sleep(0.05)

        # Custom comparator always returns True, so child should NOT re-render
        assert child_renders[0] == 1
    finally:
        _restore(saved)


def test_memo_with_hooks():
    """Memo components should still work with hooks when they do re-render."""
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        child_setter = [None]

        def Child(props):
            local, set_local = hooks.use_state(0)
            child_setter[0] = set_local
            return vdom.h("span", {}, f"local={local}")

        MemoChild = vdom.memo(Child)

        root = dom.Element(node=_Node(tag="div"))
        vdom.render(vdom.h(MemoChild, {}), root)
        assert "local=0" in _collect_texts(root.element)

        child_setter[0](5)
        time.sleep(0.05)

        assert "local=5" in _collect_texts(root.element)
    finally:
        _restore(saved)



"""Component base class for class-based VDOM components."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List

if TYPE_CHECKING:
    from .vdom import VNode

__all__ = ["Component", "forward_ref"]


class Component:
    """Class component base for the VDOM renderer.

    Subclasses should implement `render(self) -> VNode`.
    Lifecycle hooks:
      - on_mount(self)
      - on_update(self, prev_props: dict)
      - on_unmount(self)
    """

    def __init__(self, props: Dict[str, Any]) -> None:
        self.props = props
        self._cleanups: List[Callable[[], Any]] = []

    def render(self) -> "VNode":
        raise NotImplementedError

    # Lifecycle hooks
    def on_mount(self) -> None:  # noqa: D401
        """Called after the component is first mounted."""
        pass

    def on_update(self, prev_props: Dict[str, Any]) -> None:  # noqa: D401
        """Called after props update and render diff applied."""
        pass

    def on_unmount(self) -> None:  # noqa: D401
        """Called before the component is removed from the DOM."""
        pass

    # --------------- Cleanup management ---------------
    def add_cleanup(self, fn: Callable[[], Any]) -> None:
        """Register a cleanup callback to run on unmount."""
        self._cleanups.append(fn)

    # Alias for ergonomics
    on_cleanup = add_cleanup

    def _run_cleanups(self) -> None:
        """Run and clear all registered cleanup callbacks safely."""
        while self._cleanups:
            fn = self._cleanups.pop()
            try:
                fn()
            except Exception:
                # Swallow cleanup errors to avoid breaking render pipeline
                pass


def forward_ref(render_fn: Callable[..., Any]) -> Callable[..., "VNode"]:
    """Create a component that forwards a ``ref`` prop to a child element.

    The wrapped function receives ``(props, ref)`` instead of ``(props,)``,
    where *ref* is the value of the ``ref`` prop (or ``None``).

    Example::

        FancyInput = forward_ref(lambda props, ref: input_(
            type="text", ref=ref, class_name="fancy", **props,
        ))
    """

    def ForwardRefWrapper(props: Dict[str, Any]) -> "VNode":
        ref = props.get("ref")
        inner_props = {k: v for k, v in props.items() if k != "ref"}
        return render_fn(inner_props, ref)

    ForwardRefWrapper._wyb_forward_ref = True  # type: ignore[attr-defined]
    ForwardRefWrapper.__name__ = f"forward_ref({getattr(render_fn, '__name__', 'Component')})"
    ForwardRefWrapper.__qualname__ = ForwardRefWrapper.__name__
    return ForwardRefWrapper



"""Tests for create_portal – rendering children into a different container."""

import importlib
import sys
from types import ModuleType

import wybthon as _wybthon_pkg  # noqa: F401


class _ClassList:
    def __init__(self):
        self._set = set()

    def add(self, name):
        self._set.add(name)

    def remove(self, name):
        self._set.discard(name)

    def contains(self, name):
        return name in self._set


class _Style:
    def __init__(self):
        self._props = {}

    def setProperty(self, name, value):
        self._props[name] = str(value)

    def removeProperty(self, name):
        self._props.pop(name, None)


class _Node:
    def __init__(self, tag=None, text=None):
        self.tag = tag
        self.nodeValue = text
        self._is_text = text is not None
        self.parentNode = None
        self.childNodes = []
        self.attributes = {}
        self.classList = _ClassList()
        self.style = _Style()
        self.value = ""
        self.checked = False

    @property
    def nextSibling(self):
        if self.parentNode is None:
            return None
        try:
            idx = self.parentNode.childNodes.index(self)
        except ValueError:
            return None
        return self.parentNode.childNodes[idx + 1] if idx + 1 < len(self.parentNode.childNodes) else None

    def appendChild(self, node):
        if getattr(node, "parentNode", None) is not None:
            try:
                node.parentNode.childNodes.remove(node)
            except Exception:
                pass
        node.parentNode = self
        self.childNodes.append(node)
        return node

    def insertBefore(self, node, anchor):
        if getattr(node, "parentNode", None) is not None:
            try:
                node.parentNode.childNodes.remove(node)
            except Exception:
                pass
        node.parentNode = self
        if anchor is None:
            self.childNodes.append(node)
            return node
        try:
            idx = self.childNodes.index(anchor)
        except ValueError:
            self.childNodes.append(node)
            return node
        self.childNodes.insert(idx, node)
        return node

    def removeChild(self, node):
        try:
            self.childNodes.remove(node)
            node.parentNode = None
        except ValueError:
            pass
        return node

    def setAttribute(self, name, value):
        self.attributes[name] = str(value)

    def getAttribute(self, name):
        return self.attributes.get(name)

    def removeAttribute(self, name):
        self.attributes.pop(name, None)


class _Document:
    def __init__(self):
        self._listeners = {}

    def createElement(self, tag):
        return _Node(tag=tag)

    def createTextNode(self, text):
        return _Node(text=str(text))

    def addEventListener(self, event_type, proxy):
        self._listeners.setdefault(event_type, set()).add(proxy)

    def removeEventListener(self, event_type, proxy):
        s = self._listeners.get(event_type)
        if s is not None and proxy in s:
            s.remove(proxy)

    def querySelector(self, sel):
        return _Node(tag="div")

    def querySelectorAll(self, sel):
        return []


def _install_stubs():
    saved = {name: sys.modules.get(name) for name in ("js", "pyodide", "pyodide.ffi")}
    js_mod = ModuleType("js")
    js_mod.document = _Document()
    js_mod.fetch = lambda url: None
    sys.modules["js"] = js_mod

    pyodide = ModuleType("pyodide")
    ffi = ModuleType("pyodide.ffi")
    ffi.create_proxy = lambda fn: fn
    sys.modules["pyodide"] = pyodide
    sys.modules["pyodide.ffi"] = ffi
    setattr(pyodide, "ffi", ffi)
    return saved


def _restore(saved):
    for name, mod in saved.items():
        if mod is None:
            sys.modules.pop(name, None)
        else:
            sys.modules[name] = mod


def _reload_modules():
    dom = importlib.import_module("wybthon.dom")
    importlib.reload(dom)
    events = importlib.import_module("wybthon.events")
    importlib.reload(events)
    component = importlib.import_module("wybthon.component")
    importlib.reload(component)
    context = importlib.import_module("wybthon.context")
    importlib.reload(context)
    reactivity = importlib.import_module("wybthon.reactivity")
    importlib.reload(reactivity)
    hooks = importlib.import_module("wybthon.hooks")
    importlib.reload(hooks)
    vdom = importlib.import_module("wybthon.vdom")
    importlib.reload(vdom)
    return vdom, hooks, dom


def _collect_texts(node):
    out = []
    if getattr(node, "_is_text", False):
        out.append(node.nodeValue)
    for ch in getattr(node, "childNodes", []):
        out.extend(_collect_texts(ch))
    return out


def test_portal_renders_into_target_container():
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        root = dom.Element(node=_Node(tag="div"))
        portal_target = dom.Element(node=_Node(tag="div"))

        def App(props):
            return vdom.h(
                "div",
                {},
                vdom.h("p", {}, "In parent"),
                vdom.create_portal(vdom.h("p", {}, "In portal"), portal_target),
            )

        vdom.render(vdom.h(App, {}), root)

        parent_texts = _collect_texts(root.element)
        portal_texts = _collect_texts(portal_target.element)

        assert "In parent" in parent_texts
        assert "In portal" not in parent_texts
        assert "In portal" in portal_texts
    finally:
        _restore(saved)


def test_portal_renders_list_of_children():
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        root = dom.Element(node=_Node(tag="div"))
        portal_target = dom.Element(node=_Node(tag="div"))

        children = [
            vdom.h("p", {}, "Child A"),
            vdom.h("p", {}, "Child B"),
        ]

        def App(props):
            return vdom.h(
                "div",
                {},
                vdom.create_portal(children, portal_target),
            )

        vdom.render(vdom.h(App, {}), root)

        portal_texts = _collect_texts(portal_target.element)
        assert "Child A" in portal_texts
        assert "Child B" in portal_texts
    finally:
        _restore(saved)


def test_portal_unmount_cleans_up():
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        root = dom.Element(node=_Node(tag="div"))
        portal_target = dom.Element(node=_Node(tag="div"))

        def App(props):
            return vdom.h(
                "div",
                {},
                vdom.create_portal(vdom.h("span", {}, "Portal content"), portal_target),
            )

        tree = vdom.h(App, {})
        vdom.render(tree, root)

        assert "Portal content" in _collect_texts(portal_target.element)

        vdom._unmount(tree)
        assert "Portal content" not in _collect_texts(portal_target.element)
    finally:
        _restore(saved)


def test_portal_placeholder_in_parent():
    """Portal should leave an empty text placeholder in the parent container."""
    saved = _install_stubs()
    try:
        vdom, hooks, dom = _reload_modules()

        root = dom.Element(node=_Node(tag="div"))
        portal_target = dom.Element(node=_Node(tag="div"))

        def App(props):
            return vdom.h(
                "div",
                {},
                vdom.h("p", {}, "Before"),
                vdom.create_portal(vdom.h("p", {}, "Modal"), portal_target),
                vdom.h("p", {}, "After"),
            )

        vdom.render(vdom.h(App, {}), root)

        parent_texts = _collect_texts(root.element)
        assert "Before" in parent_texts
        assert "After" in parent_texts
        assert "Modal" not in parent_texts
        assert "Modal" in _collect_texts(portal_target.element)
    finally:
        _restore(saved)
